from __future__ import annotations
from cnnecharts.optionCaller import OptionCaller
import pandas as pd
from typing import TYPE_CHECKING, Optional, Callable, List
from functools import wraps

if TYPE_CHECKING:
    from cnnecharts.spec import Spec


class SeriesProp(object):
    def __init__(self) -> None:
        self.__fns: List[Callable] = []

    def _add_fns(self, *fns: Callable):
        self.__fns.extend(fns)
        return self

    def __add__(self, other: "SeriesProp"):
        newProp = SeriesProp()
        newProp._add_fns(*self.__fns, *other.__fns)

        return newProp

    def get_fns(self):
        return self.__fns


def _defineProp(fn):
    return SeriesProp()._add_fns(fn)


def id(id: str):
    def fn(series_obj: Spec):
        # series_obj["id"] = id
        if series_obj.has("name"):
            series_obj["id"] = f'{series_obj["name"]}_{id}'
        else:
            series_obj["id"] = id
        return series_obj

    return _defineProp(fn)


class dataGroupId(SeriesProp):
    """"""

    def __init__(self, id) -> None:

        super().__init__()

        def fn(series_obj):
            series_obj["dataGroupId"] = id
            return series_obj

        self._add_fns(fn)


class onUniversalTransition(SeriesProp):
    """开启动画效果"""

    def __init__(self, duration: int = 1000) -> None:

        super().__init__()

        def fn(series_obj):
            series_obj["universalTransition"] = True
            series_obj["animationDurationUpdate"] = duration
            return series_obj

        self._add_fns(fn)


class background(SeriesProp):
    def __init__(self, color: str) -> None:
        super().__init__()

        def fn(series_obj):
            series_obj["showBackground"] = True
            series_obj["backgroundStyle.color"] = color
            return series_obj

        self._add_fns(fn)


class emphasis(SeriesProp):
    def __init__(self, focus: str = "self") -> None:
        super().__init__()

        def fn(series_obj):
            series_obj["emphasis.focus"] = focus
            series_obj["emphasis.label.show"] = True
            return series_obj

        self._add_fns(fn)


class stack(SeriesProp):
    def __init__(self, stack: str = "total") -> None:
        super().__init__()

        def fn(series_obj):
            series_obj["stack"] = stack
            return series_obj

        self._add_fns(fn)


class area(SeriesProp):
    def __init__(self, smooth=True, lineWidth=0, showSymbol=False, opacity=0.8) -> None:
        super().__init__()

        def fn(series_obj):
            series_obj["smooth"] = smooth
            series_obj["lineStyle.width"] = lineWidth
            series_obj["showSymbol"] = showSymbol
            series_obj["areaStyle.opacity"] = opacity
            return series_obj

        self._add_fns(fn)


class label(SeriesProp):
    def __init__(self, formatter=r"{c}") -> None:
        """
        标签内容格式器，支持字符串模板和回调函数两种形式，字符串模板与回调函数返回的字符串均支持用 \n 换行。

        字符串模板 模板变量有：

        {a}：系列名。
        {b}：数据名。
        {c}：数据值。
        {@xxx}：数据中名为 'xxx' 的维度的值，如 {@product} 表示名为 'product' 的维度的值。
        {@[n]}：数据中维度 n 的值，如 {@[3]} 表示维度 3 的值，从 0 开始计数。
        示例：

        formatter: '{b}: {@score}'
        """
        super().__init__()

        def fn(series_obj):
            series_obj["label.show"] = True
            series_obj["label.formatter"] = formatter
            return series_obj

        self._add_fns(fn)



from __future__ import annotations
from cnnecharts.optionCaller import OptionCaller
import pandas as pd
from typing import TYPE_CHECKING, Optional, Callable, List


if TYPE_CHECKING:
    from cnnecharts.spec import Spec



import base64
from datetime import datetime
from enum import Enum
from pathlib import Path
from json import JSONEncoder, dumps
from functools import partial
import numpy as np
import pandas as pd


def _nan2None(obj):
    if isinstance(obj, dict):
        return {k: _nan2None(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_nan2None(v) for v in obj]
    elif isinstance(obj, float) and pd.isna(obj):
        return None
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, datetime):
        return str(obj)

    return obj


class _NanConverter(JSONEncoder):
    def default(self, obj):
        # possible other customizations here
        pass

    def encode(self, obj, *args, **kwargs):
        obj = _nan2None(obj)
        return super().encode(obj, *args, **kwargs)

    def iterencode(self, obj, *args, **kwargs):
        obj = _nan2None(obj)
        return super().iterencode(obj, *args, **kwargs)


json_dumps_fn = partial(dumps, cls=_NanConverter)



from __future__ import annotations
from cnnecharts.optionCaller import OptionCaller
import pandas as pd
from typing import TYPE_CHECKING, Optional

from cnnecharts.spec import Spec, OptionSpec

if TYPE_CHECKING:
    from cnnecharts.mapping import Mapping
    from cnnecharts.seriesProps import SeriesProp

from dataclasses import dataclass, field


@dataclass
class MappingData(object):
    data: Optional[pd.DataFrame] = field(init=False, default=None)
    x: Optional[str] = field(init=False, default=None)
    y: Optional[str] = field(init=False, default=None)
    value: Optional[str] = field(init=False, default=None)
    color: Optional[str] = field(init=False, default=None)


class ChartPart(OptionCaller):
    def __init__(self, *props: SeriesProp) -> None:
        super().__init__()
        self.series_callers = []
        self._mappingData = MappingData()

        for p in props:
            self.series_callers.extend(p.get_fns())

    def mapping(
        self,
        *,
        data: Optional[pd.DataFrame] = None,
        x: Optional[str] = None,
        y: Optional[str] = None,
        value: Optional[str] = None,
        color: Optional[str] = None,
    ):
        self._mappingData.data = data
        self._mappingData.x = x
        self._mappingData.y = y
        self._mappingData.value = value
        self._mappingData.color = color

        return self


class Bar(ChartPart):
    def __init__(self, *props: SeriesProp) -> None:
        super().__init__(*props)

    def mapping(
        self,
        *,
        data: Optional[pd.DataFrame] = None,
        x: Optional[str] = None,
        y: Optional[str] = None,
        color: Optional[str] = None,
    ):
        return super().mapping(data=data, x=x, y=y, color=color)

    def _ex_create_spec(self, mapping: Mapping, spec: OptionSpec):
        data = self._mappingData.data
        x = self._mappingData.x
        y = self._mappingData.y
        color = self._mappingData.color

        data, x, y, color = mapping.transform(data, x, y, color)
        data = data.round(
            2,
        )

        xAxis = spec["xAxis"]
        xAxis["type"] = "category"
        xAxis["data"] = data.index.tolist()

        spec["yAxis.type"] = "value"

        series = [
            {
                "type": "bar",
                "name": col,
                "data": list(data[col]),
            }
            for col in data.columns
        ]

        for series_obj in series:
            for caller in self.series_callers:
                series_obj = caller(Spec(series_obj))

            if (
                isinstance(series_obj["universalTransition"], bool)
                and series_obj["universalTransition"]
            ):
                series_obj["id"] = series_obj["name"]
                series_obj["dataGroupId"] = series_obj["name"]

            spec.add_series(series_obj)

        # spec["legend.data"] = list(data.columns)

        return spec


class Line(ChartPart):
    def __init__(self, *props: SeriesProp) -> None:
        super().__init__(*props)

    def mapping(
        self,
        *,
        data: Optional[pd.DataFrame] = None,
        x: Optional[str] = None,
        y: Optional[str] = None,
        color: Optional[str] = None,
    ):
        return super().mapping(data=data, x=x, y=y, color=color)

    def _ex_create_spec(self, mapping: Mapping, spec: OptionSpec):
        data = self._mappingData.data
        x = self._mappingData.x
        y = self._mappingData.y
        color = self._mappingData.color

        data, x, y, color = mapping.transform(data, x, y, color)

        xAxis = spec["xAxis"]
        xAxis["type"] = "category"
        xAxis["data"] = data.index.tolist()

        spec["yAxis.type"] = "value"

        series = [
            {
                "type": "line",
                "name": col,
                "data": list(data[col]),
            }
            for col in data.columns
        ]

        for series_obj in series:
            for caller in self.series_callers:
                series_obj = caller(Spec(series_obj))

            spec.add_series(series_obj)

            if (
                isinstance(series_obj["universalTransition"], bool)
                and series_obj["universalTransition"]
            ):
                series_obj["id"] = series_obj["name"]
                series_obj["dataGroupId"] = series_obj["name"]

        spec["legend.data"] = list(data.columns)

        return spec


class Pie(ChartPart):
    def __init__(self, *props: SeriesProp) -> None:
        super().__init__(*props)

    def mapping(
        self,
        *,
        data: Optional[pd.DataFrame] = None,
        value: Optional[str] = None,
        color: Optional[str] = None,
    ):
        return super().mapping(data=data, value=value, color=color)

    def _ex_create_spec(self, mapping: Mapping, spec: OptionSpec):
        data = self._mappingData.data or mapping.data
        value = self._mappingData.value or mapping.value
        color = self._mappingData.color or mapping.color

        data = data.groupby(color)[value].mean()

        spec["xAxis"] = {}

        spec["yAxis"] = {}

        # TODO: 应该依据是否开启动画设置 groupId
        series_data = [
            {"value": v, "name": n, "groupId": n}
            for n, v in zip(data.index.values, data.values)
        ]

        series = [
            {
                "type": "pie",
                "name": color,
                "data": series_data,
            }
        ]

        for series_obj in series:
            for caller in self.series_callers:
                series_obj = caller(Spec(series_obj))

            if (
                isinstance(series_obj["universalTransition"], bool)
                and series_obj["universalTransition"]
            ):
                

                series_obj["universalTransition"] = {
                    "enabled": True,
                    "seriesKey": data.index.values[0] if len(data.index.values)==1 else list(data.index.values),
                }
            spec.add_series(series_obj)

        spec["legend"] = {}

        return spec



from typing import Dict, List, Any, Optional
import json


class Spec:
    def __init__(self, init_data: Optional[Dict] = None) -> None:
        data = {} if init_data is None else init_data
        if isinstance(init_data, Spec):
            data = init_data._data

        self._data = data

    def has(self, key: str):
        return key in self._data

    def __getitem__(self, key: str) -> "Spec":
        paths = key.split(".")

        data = self._data
        for p in paths:

            if p not in data:
                data[p] = {}
            data = data[p]

        if isinstance(data, dict):
            return Spec(data)
        return data

    def __setitem__(self, key: str, value: Any):
        paths = key.split(".")

        data = self._data
        for p in paths[:-1]:
            if p not in data:
                data[p] = {}
            data = data[p]

        if isinstance(value, Spec):
            value = value._data

        data[paths[-1]] = value

    def __delitem__(self, key: str):
        paths = key.split(".")

        data = self._data
        for p in paths[:-1]:
            if p not in data:
                data[p] = {}
            data = data[p]

        del data[paths[-1]]

    def __str__(self) -> str:
        return json.dumps(self._data, indent=2)

    def __repr__(self) -> str:
        return str(self)


class OptionSpec(Spec):
    def __init__(self, init_data: Optional[Dict] = None) -> None:
        super().__init__(init_data)
        self["series"] = []

    def add_series(self, obj: Dict):
        if isinstance(obj, Spec):
            obj = obj._data
        self._data["series"].append(obj)

    # @property
    # def series(self) -> List[Any]:
    #     """The series property."""
    #     if "series" not in self._data:
    #         self["series"] = []

    #     return self["series"]



from jinja2 import PackageLoader, Environment
from typing import TYPE_CHECKING, Dict, Optional
from cnnecharts.utils.data_gen import json_dumps_fn
from cnnecharts.utils.baseInfos import get_root_dir

env = Environment(loader=PackageLoader("cnnecharts", "templates"))


def htmlRender(options_content: str, theme_json: Optional[str] = None):
    template = env.get_template("fileRender.html")

    if theme_json is None:
        theme_path = get_root_dir() / "templates/themes/macarons.json"
        with open(theme_path, mode="r", encoding="utf8") as f:
            theme_json = f.read()

    return template.render(options=options_content, theme=theme_json)
